#include "int_packer.h"

#include "globals.h"
#include "painting.h"

#include <algorithm>
#include <cassert>
#include <iostream>

using namespace std;


static const int BITS_PER_BIN = sizeof(IntPacker::Bin) * 8;


static IntPacker::Bin get_bit_mask(int from, int to) {
    // Return mask with all bits in the range [from, to) set to 1.
    assert(from >= 0 && to >= from && to <= BITS_PER_BIN);
    int length = to - from;
    if (length == BITS_PER_BIN) {
        // 1U << BITS_PER_BIN has undefined behaviour in C++; e.g.
        // 1U << 32 == 1 (not 0) on 32-bit Intel platforms. Hence this
        // special case.
        assert(from == 0 && to == BITS_PER_BIN);
        return ~IntPacker::Bin(0);
    } else {
        return ((IntPacker::Bin(1) << length) - 1) << from;
    }
}

static int get_bit_size_for_range(unsigned int range) {
    int num_bits = 0;
    while ((1U << num_bits) < range)
        ++num_bits;
    return num_bits;
}

class IntPacker::VariableInfo {
    unsigned int range;
    int bin_index;
    int shift;
    Bin read_mask;
    Bin clear_mask;
public:
    VariableInfo(unsigned int range_, int bin_index_, int shift_)
        : range(range_),
          bin_index(bin_index_),
          shift(shift_) {
        if (range == pow(2, 32) - 1){
            clear_mask = 0;
        } else {
            int bit_size = get_bit_size_for_range(range);
            read_mask = get_bit_mask(shift, shift + bit_size);
            clear_mask = ~read_mask;
        }
    }

    VariableInfo()
        : bin_index(-1), shift(0), read_mask(0), clear_mask(0) {
        // Default constructor needed for resize() in pack_bins.
    }

    ~VariableInfo() {
    }

    int get(const Bin *buffer) const {
        return (buffer[bin_index] & read_mask) >> shift;
    }

    void set(Bin *buffer, int value) const {
        assert(value >= 0 && ((unsigned int) value) < range);
        Bin &bin = buffer[bin_index];
        bin = (bin & clear_mask) | (value << shift);
    }
    
    bool get_bit(const Bin *buffer, int value) const {
        return (buffer[bin_index] & 1 << (shift + value)) >> (shift + value);
    }

    void set_bit(Bin *buffer, int value) const {
        assert(value >= 0 && pow(2, value) < range);
        Bin &bin = buffer[bin_index];
        bin = bin | (1 << (shift + value));
    }
    
    void init_zero(Bin *buffer) const {
        Bin &bin = buffer[bin_index];
        bin = (bin & clear_mask);
    }
};


IntPacker::IntPacker(const vector<int> &ranges)
    : num_bins(0),
    additional_bins(0) {
        
    if (g_use_redblack_search){
        var_to_bin.resize(g_variable_domain.size(), -1);
    }
    pack_bins(ranges);
}

IntPacker::~IntPacker() {
}

int IntPacker::get(const Bin *buffer, int var) const {
    return var_infos[var].get(buffer);
}

void IntPacker::set(Bin *buffer, int var, int value) const {
    var_infos[var].set(buffer, value);
}

bool IntPacker::get_bit(const Bin *buffer, int var, int value) const {
    if (value >= 32){
        var = var_to_bin[var];
        value -= 32;
        while (value >= 32){
            ++var;
            value -= 32;
        }
    }
    return var_infos[var].get_bit(buffer, value);
}

void IntPacker::set_bit(Bin *buffer, int var, int value) const {
    if (value >= 32){
        var = var_to_bin[var];
        value -= 32;
        while (value >= 32){
            ++var;
            value -= 32;
        }
    }
    var_infos[var].set_bit(buffer, value);
}

void IntPacker::init_zero(IntPacker::Bin* buffer, int var) const {
    var_infos[var].init_zero(buffer);
    int i = 0;
    int number_bits = 32;
    while (number_bits < g_variable_domain[var]){
        var_infos[var_to_bin[var] + i].init_zero(buffer);
        number_bits += 32;
        ++i;
    }
}

void IntPacker::pack_bins(const vector<int> &ranges) {
    assert(var_infos.empty());

    int num_vars = ranges.size();
    var_infos.resize(num_vars);

    // bits_to_vars[k] contains all variables that require exactly k
    // bits to encode. Once a variable is packed into a bin, it is
    // removed from this index.
    // Loop over the variables in reverse order to prefer variables with
    // low indices in case of ties. This might increase cache-locality.
    vector<vector<int> > bits_to_vars(BITS_PER_BIN + 1);
    for (int var = num_vars - 1; var >= 0; --var) {
        int bits = get_bit_size_for_range(ranges[var]);
        if (g_use_redblack_search && g_painting->is_red_var(var)) {
            bits = g_variable_domain[var];
            if (bits >= (int) bits_to_vars.size()){
                bits_to_vars.resize(bits + 1);
            }
        }
        assert((g_use_redblack_search && g_painting->is_red_var(var)) || bits <= BITS_PER_BIN);
        bits_to_vars[bits].push_back(var);
    }

    int packed_vars = 0;
    while (packed_vars != num_vars) {
        packed_vars += pack_one_bin(ranges, bits_to_vars);
    }
}

int IntPacker::pack_one_bin(const vector<int> &ranges,
                            vector<vector<int> > &bits_to_vars) {
    // Returns the number of variables added to the bin. We pack each
    // bin with a greedy strategy, always adding the largest variable
    // that still fits.

    ++num_bins;
    int bin_index = num_bins - 1;
    int used_bits = 0;
    int num_vars_in_bin = 0;

    while (true) {
        // Determine size of largest variable that still fits into the bin.
        int bits = used_bits == 0 ? bits_to_vars.size() - 1 : BITS_PER_BIN - used_bits;
        while (bits > 0 && bits_to_vars[bits].empty())
            --bits;

        if (bits <= 0) {
            // No more variables fit into the bin.
            // (This also happens when all variables have been packed.)
            return num_vars_in_bin;
        }

        // We can pack another variable of size bits into the current bin.
        // Remove the variable from bits_to_vars and add it to the bin.
        vector<int> &best_fit_vars = bits_to_vars[bits];
        int var = best_fit_vars.back();
        best_fit_vars.pop_back();

        if (g_use_redblack_search && g_painting->is_red_var(var)){
            int stored_bits = 0;
            while (stored_bits < bits){
                unsigned int range;
                if (bits - stored_bits >= 32){
                    range = pow(2, 32) - 1;
                } else {
                    range = pow(2, bits - stored_bits);
                }
                if (stored_bits == 0){
                    var_infos[var] = VariableInfo(range, bin_index, used_bits);
                    if (bits > 32){
                        var_to_bin[var] = g_variable_domain.size() + additional_bins;
                    }
                    stored_bits += min(32, bits - stored_bits);
                } else {
                    ++num_bins;
                    ++additional_bins;

                    var_infos.resize(g_variable_domain.size() + additional_bins);
                    var_infos[var_infos.size() - 1] = VariableInfo(range, num_bins - 1, used_bits);
                    
                    stored_bits += min(32, bits - stored_bits);
                }
            }
        } else {
            var_infos[var] = VariableInfo(ranges[var], bin_index, used_bits);
        }
        used_bits += bits;
        ++num_vars_in_bin;
    }
}
