#include "state_registry.h"

#include "axioms.h"
#include "operator.h"
#include "successor_generator.h"
#include "painting.h"
#include "per_state_information.h"

#include <limits>
#include <stdint.h>

using namespace std;

StateRegistry::StateRegistry()
    : state_data_pool(g_state_packer->get_num_bins()),
      registered_states(0,
                        StateIDSemanticHash(state_data_pool),
                        StateIDSemanticEqual(state_data_pool)),
      cached_initial_state(0) {
}

StateRegistry::~StateRegistry() {
    for (set<PerStateInformationBase *>::iterator it = subscribers.begin();
         it != subscribers.end(); ++it) {
        (*it)->remove_state_registry(this);
    }
    delete cached_initial_state;
}

StateID StateRegistry::insert_id_or_pop_state() {
    /*
      Attempt to insert a StateID for the last state of state_data_pool
      if none is present yet. If this fails (another entry for this state
      is present), we have to remove the duplicate entry from the
      state data pool.
    */
    StateID id(state_data_pool.size() - 1);
    pair<StateIDSet::iterator, bool> result = registered_states.insert(id);
    bool is_new_entry = result.second;
    if (!is_new_entry) {
        state_data_pool.pop_back();
    }
    assert(registered_states.size() == state_data_pool.size());
    return *result.first;
}

State StateRegistry::lookup_state(StateID id) const {
    return State(state_data_pool[id.value], *this, id);
}

const State &StateRegistry::get_initial_state() {
    if (cached_initial_state == 0) {
        PackedStateBin *buffer = new PackedStateBin[g_state_packer->get_num_bins()];
        for (size_t var = 0; var < g_initial_state_data.size(); ++var) {
            if (g_use_redblack_search && g_painting->is_red_var(var)){
                g_state_packer->init_zero(buffer, var);
                g_state_packer->set_bit(buffer, var, g_initial_state_data[var]);
            } else {
                g_state_packer->set(buffer, var, g_initial_state_data[var]);
            }
        }
        
        if (g_use_redblack_search){
            saturate_state(buffer, true);
            init_state_best_supporters = get_stored_best_supporters();
        }
        
        g_axiom_evaluator->evaluate(buffer);
        
        state_data_pool.push_back(buffer);
        // buffer is copied by push_back
        delete[] buffer;
        
        StateID id = insert_id_or_pop_state();
        cached_initial_state = new State(lookup_state(id));
        
#ifdef DEBUG_SEARCH
        cout << "dump init state StateRegistry" << endl;
        cached_initial_state->dump_pddl();
#endif
    }
    return *cached_initial_state;
}

//TODO it would be nice to move the actual state creation (and operator application)
//     out of the StateRegistry. This could for example be done by global functions
//     operating on state buffers (PackedStateBin *).
State StateRegistry::get_successor_state(const State &predecessor, const Operator &op, bool store_best_supporters) {    
    assert(op.is_applicable(predecessor));        
    assert(!op.is_axiom());
    assert(!g_use_redblack_search || op.is_black_op());
    
#ifdef DEBUG_SEARCH
    if (store_best_supporters){
        predecessor.dump_pddl();
        cout << op.get_name() << endl;
    }
#endif
    
    state_data_pool.push_back(predecessor.get_packed_buffer());
    PackedStateBin *buffer = state_data_pool[state_data_pool.size() - 1];

    for (size_t i = 0; i < op.get_effects().size(); ++i) {
        const Effect &effect = op.get_effects()[i];
        if (g_use_redblack_search && g_painting->is_red_var(effect.var)){
            g_state_packer->set_bit(buffer, effect.var, effect.val);
        } else if (effect.does_fire(predecessor)){
            g_state_packer->set(buffer, effect.var, effect.val);
        }
    }
    if (g_use_redblack_search){
        saturate_state(buffer, store_best_supporters);
    }
    
    g_axiom_evaluator->evaluate(buffer);
    
    StateID id = insert_id_or_pop_state();
    
    return lookup_state(id);
}

void StateRegistry::saturate_state(PackedStateBin *buffer, bool store_best_supporters) const {
    
    vector<vector<int> > lowest_cost(g_variable_domain.size());

    if (store_best_supporters){
        cached_best_supporters.resize(g_variable_domain.size());
    }
    for (size_t var = 0; var < g_variable_domain.size(); ++var){
        if (g_painting->is_red_var(var)){
            lowest_cost[var].resize(g_variable_domain[var], -1);
            for (int val = 0; val < g_variable_domain[var]; ++val){
                if (g_state_packer->get_bit(buffer, var, val)){
                    lowest_cost[var][val] = 0;
                }
            }
            if (store_best_supporters){
                vector<const Operator*>(g_variable_domain[var]).swap(cached_best_supporters[var]);
            }
        }       
    }

    bool change = true;
    while (change){
        change = false;
        
        vector<const Operator*> applicable_ops;
        g_successor_generator->generate_applicable_ops(buffer, applicable_ops);
        
        for (auto const &op : applicable_ops){
#ifndef NDEBUG
            bool is_applicable = true;
            for (auto const &pre : op->get_black_pre()){
                if (g_state_packer->get(buffer, pre->var) != pre->val){
                    is_applicable = false;
                    break;
                }
            }
            assert(is_applicable);
#endif
            bool changes_black = false;
            if (op->is_black_op()){
                const vector<const Effect*> &black_effs = op->get_black_eff();
                for (size_t eff = 0; eff < black_effs.size(); ++eff){
                    int var = black_effs[eff]->var;
                    int val = black_effs[eff]->val;
                    
                    if (val != g_state_packer->get(buffer, var)){
                        changes_black = true;
                        break;
                    }
                }
            }
            if (!changes_black){
                int max_pre_cost = 0;
                const vector<const Condition*> &red_pre = op->get_red_pre();
                for (size_t pre = 0; pre < red_pre.size(); ++pre){
                    int var = red_pre[pre]->var;
                    int val = red_pre[pre]->val;
                                       
                    assert(lowest_cost[var][val] >= 0);
                    max_pre_cost = max(max_pre_cost, lowest_cost[var][val]);
                }
                
                max_pre_cost += get_adjusted_action_cost(*op, cost_type);
                
                const vector<const Effect*> &red_effs = op->get_red_eff();
                for (size_t eff = 0; eff < red_effs.size(); ++eff){
                    int var = red_effs[eff]->var;
                    int val = red_effs[eff]->val;
                    // TODO only need to check for cheaper paths when storing best supporters
                    if (lowest_cost[var][val] == -1 || max_pre_cost < lowest_cost[var][val]){
                        g_state_packer->set_bit(buffer, var, val);
                        lowest_cost[var][val] = max_pre_cost;
                        change = true;
                        if (store_best_supporters){
                            cached_best_supporters[var][val] = op;
                        }
                    }
                }
            }
        }
    }
}

void StateRegistry::subscribe(PerStateInformationBase *psi) const {
    subscribers.insert(psi);
}

void StateRegistry::unsubscribe(PerStateInformationBase *const psi) const {
    subscribers.erase(psi);
}
